﻿--------------------------------------------
-- Intro to spatial analysis with PostGIS 
--------------------------------------------
-- Start the PostgreSQL service
-- Start PgAdmin
-- Connect to the service
-- Create a new database
-- Install PostGIS in the new database and check version
CREATE EXTENSION postgis;
SELECT postgis_full_version();


-- Load the forest cover shapefile
 -- Determine the SRID with http://prj2epsg.org/search
 -- Search in the spatial_ref_sys table
SELECT * FROM spatial_ref_sys WHERE srtext LIKE '%MTM%' AND srtext LIKE '%NAD83%';

 -- Load the first forest cover shapefile
-- shp2pgsql -s 32187 -W LATIN1 -I "C:\DATA\Formations\PostGIS\AtelierGISDay2014\data\FM\couvertforestier\03\c08peefo.shp" "a_forestcover_mtm7" | psql -U "postgres" -d "intropostgis"

 -- Display in OpenJump
SELECT * FROM a_forestcover_mtm7;

 -- Load the second shapefile
-- shp2pgsql -a -s 32187 -W LATIN1 "C:\DATA\Formations\PostGIS\AtelierGISDay2014\data\FM\couvertforestier\06\c08peefo.shp" "a_forestcover_mtm7" | psql -U "postgres" -d "intropostgis"

-- Load the Foret Montmorency limits
-- shp2pgsql -s 4269:32187 -W LATIN1 -I "C:\DATA\Formations\PostGIS\AtelierGISDay2014\data\FM\limites\limites_FM.shp" "b_limites_fm_mtm7" | psql -U "postgres" -d "intropostgis"

-- Check in the geometry_column view that limites is in 32187
-- Display in OpenJump
SELECT * FROM b_limites_fm_mtm7;

-- Install PostGIS Addons

--------------------------------------------
-- 1) Extraction from POLYGONS for POINTS 
--------------------------------------------
-- Generate random points in the limites_fm
CREATE TABLE c_random_points_fm_1000_mtm7 AS
SELECT ST_RandomPoints(ST_Union(geom), 1000) geom FROM b_limites_fm_mtm7;

-- Add a spatial index on the points
CREATE INDEX c_random_points_fm_1000_mtm7_geom_gist ON c_random_points_fm_1000_mtm7 USING gist (geom);

-- Add a unique identifier to each point
SELECT ST_AddUniqueID('c_random_points_fm_1000_mtm7', 'id', true);

-- Display in OpenJump
SELECT * FROM c_random_points_fm_1000_mtm7;

-- Extract a cover type value for each POINT
CREATE TABLE c_random_points_fm_1000_cover_mtm7 AS
SELECT p.id, f.ges_co, p.geom
FROM c_random_points_fm_1000_mtm7 p, a_forestcover_mtm7 f
WHERE ST_Intersects(p.geom, f.geom);

-- Display in OpenJump
SELECT * FROM c_random_points_fm_1000_cover_mtm7;

--------------------------------------------
-- 2) Extraction from POLYGONS for POLYGONS 
--------------------------------------------
-- Create a 100m buffer table and index it. 30000 m square
CREATE TABLE d_random_buffers_fm_1000_mtm7 AS
SELECT id, ST_Buffer(geom, 100) geom
FROM c_random_points_fm_1000_mtm7;

-- Create the a spatial index on the buffers
CREATE INDEX d_random_buffers_fm_1000_mtm7_geom_gist ON d_random_buffers_fm_1000_mtm7 USING gist (geom);

-- Display in OpenJump
SELECT * FROM d_random_buffers_fm_1000_mtm7;

--------------------------------------------
-- 2.1) Extraction from POLYGONS for POLYGONS (areas)
--------------------------------------------
-- Extract the area covered by each type of forest for each buffer. 11s
CREATE TABLE d_random_buffers_fm_coverarea_1000_mtm7 AS
SELECT id, ges_co, ST_Area(geom) area, geom
FROM (SELECT ST_Intersection(o.geom, c.geom) geom, id, ges_co 
      FROM d_random_buffers_fm_1000_mtm7 o, a_forestcover_mtm7 c
      WHERE ST_Intersects(o.geom, c.geom)) foo;

-- Display
SELECT * FROM d_random_buffers_fm_coverarea_1000_mtm7;

--------------------------------------------
-- 2.1) Extraction from POLYGONS for POLYGONS (proportions)
--------------------------------------------
-- Aggregate per cover type and compute proportions. 11s.
CREATE TABLE d_random_buffers_fm_coverprop_1000_mtm7 AS
SELECT id, ges_co, sum(area) / min(bufferarea) prop , ST_Union(geom) geom
FROM (SELECT id, ges_co, ST_Area(geom) area, bufferarea, geom
      FROM (SELECT ST_Intersection(o.geom, c.geom) geom, id, ges_co, ST_Area(o.geom) bufferarea
            FROM d_random_buffers_fm_1000_mtm7 o, a_forestcover_mtm7 c
            WHERE ST_Intersects(o.geom, c.geom)
           ) foo
     ) foo2
GROUP BY id, ges_co
ORDER BY id, prop DESC;

-- Display
SELECT * FROM d_random_buffers_fm_coverprop_1000_mtm7;

--------------------------------------------
-- 3) Extraction from RASTER
--------------------------------------------
-- Load elevation raster files covering the Foret Montmorency in-db. 4m
-- raster2pgsql -t 20x20 -I -C -x C:\DATA\Formations\PostGIS\AtelierGISDay2014\data\quebec\srtm\tif/srtm_22_03.tif e_elevation_fm_20x20_wgs84 | psql -U postgres -d intropostgis

-- raster2pgsql -t 10x10 -I -C -x C:\DATA\Formations\PostGIS\AtelierGISDay2014\data\quebec\srtm\tif/srtm_22_03.tif e_elevation_fm_10x10_wgs84 | psql -U postgres -d intropostgis

-- Retile the 100x100 to 10x10. XXs
CREATE TABLE e_elevation_fm_10x10_wgs84_out AS
SELECT generate_series(1, 360000) id, ST_Tile(rast, 10, 10) rast
FROM e_elevation_fm_100x100_wgs84_out;

CREATE TABLE e_elevation_fm_20x20_wgs84_out AS
SELECT generate_series(1, 90000) id, ST_Tile(rast, 20, 20) rast
FROM e_elevation_fm_100x100_wgs84_out;

-- Index it. 14s
CREATE INDEX e_elevation_fm_10x10_wgs84_rast_gist ON e_elevation_fm_10x10_wgs84 USING gist (st_convexhull(rast));

-- Add a unique identifier to each point if not certain about the number to put in generate_series(). 60s
SELECT ST_AddUniqueID('e_elevation_fm_10x10_wgs84', 'rid', true);

-- Add all constraints except extent and regular_blocking. 60s
SELECT AddRasterConstraints('e_elevation_fm_10x10_wgs84', 'rast', true, true, true, true, true, true, false, true, true, true, true, false);

-- Display the tiles in OpenJump
SELECT rid, rast::geometry FROM e_elevation_fm_20x20_wgs84;
SELECT ST_Transform(geom, 4326) geom FROM b_limites_fm_mtm7;

-- Display one tile
SELECT (ST_DumpAsPolygons(rast)).*
FROM e_elevation_fm_20x20_wgs84
WHERE rid = 48233;

-- Compute size of biggest tables
SELECT nspname || '.' || relname AS "relation",
       pg_size_pretty(pg_total_relation_size(C.oid)) AS "total_size"
FROM pg_class C
LEFT JOIN pg_namespace N ON (N.oid = C.relnamespace)
WHERE nspname NOT IN ('pg_catalog', 'information_schema')
      AND C.relkind <> 'i'
      AND nspname !~ '^pg_toast'
ORDER BY pg_total_relation_size(C.oid) DESC
LIMIT 50;

--------------------------------------------
-- 3.1) Extraction from RASTER for POINTS 
--------------------------------------------
-- Extract the elevation for each point. 3s
CREATE TABLE f_random_points_fm_1000_elev_mtm7 AS
SELECT id, geom, ST_Value(rast, ST_Transform(geom, 4326)) elev
FROM c_random_points_fm_1000_mtm7, e_elevation_fm_10x10_wgs84
WHERE ST_Intersects(rast, ST_Transform(geom, 4326));

-- Display in OpenJump
SELECT id, elev, ST_Transform(geom, 4326) geom
FROM f_random_points_fm_1000_elev_mtm7

------------------------------------------------------
-- 3.2) Extraction from RASTER for POLYGONS, vector mode
------------------------------------------------------
-- Intersect each buffer with the elevation in vector mode. 38s
CREATE TABLE f_random_buffers_fm_1000_elev_mtm7 AS
SELECT id, (gv).val elev, ST_Transform((gv).geom,32187) geom, ST_Area(ST_Transform((gv).geom, 32187)) area
FROM (SELECT id, ST_Intersection(rast, ST_Transform(geom, 4326)) gv
      FROM d_random_buffers_fm_1000_mtm7, e_elevation_fm_20x20_wgs84
      WHERE ST_Intersects(rast, ST_Transform(geom, 4326))
     ) foo;

-- Display as table
SELECT * FROM f_random_buffers_fm_1000_elev_mtm7;
SELECT id, elev, ST_Transform(geom,4326) FROM f_random_buffers_fm_1000_elev_mtm7;

-- Display in OpenJump
SELECT id, elev, ST_Transform(geom, 4326) geom, area
FROM f_random_buffers_fm_1000_elev_mtm7;

-- Intersect and aggregate per buffer. 32s
CREATE TABLE f_random_buffers_fm_1000_stats_mtm7 AS
SELECT id, (ST_AreaWeightedSummaryStats(gv)).*
FROM (SELECT id, ST_Intersection(rast, ST_Transform(geom, 4326)) gv
      FROM d_random_buffers_fm_1000_mtm7, e_elevation_fm_10x10_wgs84
      WHERE ST_Intersects(rast, ST_Transform(geom, 4326))
     ) foo
GROUP BY id;

-- Display as table
SELECT * FROM f_random_buffers_fm_1000_stats_mtm7;

-- Display in OpenJump
SELECT id, weightedmean, ST_Transform(geom, 4326) geom
FROM f_random_buffers_fm_1000_stats_mtm7;

------------------------------------------------------
-- 3.3) Extraction from RASTER for POLYGONS, raster mode
------------------------------------------------------
-- Load elevation raster files covering the province out-db 
-- raster2pgsql -R -F -I -C -x C:\DATA\Formations\PostGIS\AtelierGISDay2014\data\quebec\srtm\tif/*.tif g_elevation_qc_wgs84_out | psql -U postgres -d intropostgis

-- Retile to 100x100. 5s
CREATE TABLE g_elevation_qc_100x100_wgs84_out AS
SELECT ST_Tile(rast, 100, 100) rast
FROM g_elevation_qc_wgs84_out;

-- Index it.
CREATE INDEX g_elevation_qc_100x100_wgs84_out_rast_gist ON g_elevation_qc_100x100_wgs84_out USING gist (st_convexhull(rast));

-- Add a unique identifier to each point. 3s
SELECT ST_AddUniqueID('g_elevation_qc_100x100_wgs84_out', 'rid', true);

-- Add constraints. 60s
SELECT AddRasterConstraints('g_elevation_qc_100x100_wgs84_out', 'rast', true, true, true, true, true, true, false, true, true, true, true, false);

-- Display the tiles in OpenJump
SELECT rid, rast::geometry FROM g_elevation_qc_100x100_wgs84_out;

-- Load province limites
-- shp2pgsql -s 4269 -W LATIN1 -I "C:\DATA\Formations\PostGIS\AtelierGISDay2014\data\quebec\limites\province.shp" "h_limites_qc_nad83" | psql -U "postgres" -d "intropostgis"

-- Display
SELECT * FROM h_limites_qc_nad83;

-- Display one tile
SELECT (ST_DumpAsPolygons(rast)).* FROM g_elevation_qc_100x100_wgs84_out WHERE rid = 20325;

-- Create 1000 2000km buffers in the province. 5s
CREATE TABLE i_random_buffers_qc_1000_wgs84 AS
SELECT ST_Transform(ST_Buffer(ST_Transform(ST_RandomPoints(ST_Union(geom), 1000), 32187), 2000), 4326) geom
FROM h_limites_qc_nad83;

CREATE INDEX i_random_buffers_qc_1000_wgs84_geom_gist ON i_random_buffers_qc_1000_wgs84 USING gist (geom);

-- Add a unique identifier to each point
SELECT ST_AddUniqueID('i_random_buffers_qc_1000_wgs84', 'id', true);

-- Display in OpenJump
SELECT * FROM i_random_buffers_qc_1000_wgs84;

------------------------------------------------------
-- Actual query
------------------------------------------------------
-- Clip and summarise the elevation tiles for each buffer. 33s
CREATE TABLE i_random_buffers_qc_1000_elevstats_wgs84 AS
SELECT id, ST_Union(geom) geom, (ST_SummaryStatsAgg(ST_Clip(rast, geom, true))).*
FROM g_elevation_qc_100x100_wgs84_out, i_random_buffers_qc_1000_wgs84
WHERE ST_Intersects(rast, geom)
GROUP BY id;

-- Display as table
SELECT * FROM i_random_buffers_qc_1000_elevstats_wgs84;

-- Display the clipping of the tiles in OpenJump. 17s
SELECT id, ST_Clip(rast, geom, true)::geometry
FROM g_elevation_qc_100x100_wgs84_out, i_random_buffers_qc_1000_wgs84
WHERE ST_Intersects(rast, geom);

------------------------------------------------------
-- 3.4) Elevation profile
------------------------------------------------------
-- Load the road network
-- shp2pgsql -s 4269:32187 -W LATIN1 -I "C:\DATA\Formations\PostGIS\AtelierGISDay2014\data\FM\routes\roads.shp" "l_routes_fm_mtm7" | psql -U "postgres" -d "intropostgis"

-- Add a unique identifier to each route. 1s
SELECT ST_AddUniqueID('l_routes_fm_mtm7', 'id', true);

-- Add an index on id
CREATE INDEX l_routes_fm_mtm7_id ON l_routes_fm_mtm7 (id);

-- Display them in OpenJump
SELECT * FROM l_routes_fm_mtm7;

-- Select the longest route
SELECT id, ST_Length(geom) length
FROM l_routes_fm_mtm7
ORDER BY length DESC;

-- Display it in mtm7
SELECT id, geom 
FROM l_routes_fm_mtm7 
WHERE id = 2000;

-- Display it in wgs84
SELECT id, ST_Transform(geom, 4326) geom 
FROM l_routes_fm_mtm7 
WHERE id = 2000;

-- Create 100 points along the route
SELECT ST_Transform(ST_LineInterpolatePoint(geom, id/100.0), 4326) geom
FROM generate_series(0, 100) id, (SELECT (ST_Dump(geom)).geom 
FROM l_routes_fm_mtm7 WHERE id = 2000) foo;

-- Extract the elevation for each point
CREATE TABLE g_elevation_profile_mtm7 AS
WITH points AS (
SELECT id, ST_LineInterpolatePoint(geom, id/100.0) geom
FROM generate_series(0, 100) id, (SELECT (ST_Dump(geom)).geom FROM l_routes_fm_mtm7 WHERE id = 2000) foo
)
SELECT id, geom, ST_Value(rast, ST_Transform(geom, 4326)) elev
FROM points, e_elevation_fm_10x10_wgs84
WHERE ST_Intersects(rast, ST_Transform(geom, 4326));

-- Display in OpenJump
SELECT id, elev, ST_Transform(geom, 4326) geom 
FROM g_elevation_profile_mtm7

-- Export in CSV
COPY g_elevation_profile_mtm7 TO '/temp/elevation_profile.csv' DELIMITER ',' CSV HEADER;
------------------------------------------------------
-- 4) Nearest neighbour
------------------------------------------------------
-- Generate random points in the limites_fm. 40s
CREATE TABLE k_random_points_fm_1000000_mtm7 AS
SELECT generate_series(1, 1000000) id, ST_RandomPoints(ST_Union(geom), 1000000) geom FROM b_limites_fm_mtm7;

-- Add a spatial index on the points. 40s
CREATE INDEX k_random_points_fm_1000000_mtm7_geom_gist ON k_random_points_fm_1000000_mtm7 USING gist (geom);

-- Add an index on id. 5s
CREATE INDEX k_random_points_fm_1000000_mtm7_id ON k_random_points_fm_1000000_mtm7 (id);

-- Display in OpenJump
SELECT * FROM k_random_points_fm_1000000_mtm7;

------------------------------------------------------
-- 4.1) Nearest neighbour - For 1 POINT from 1 000 000 POINTS. Classic method with and without ST_DWithin.
------------------------------------------------------
-- Without ST_DWithin()
SELECT pointB.geom, pointB.id, ST_Distance(pointA.geom, pointB.geom) dist
FROM c_random_points_fm_1000_mtm7 pointA, k_random_points_fm_1000000_mtm7 pointB
WHERE pointA.id = 146 
ORDER BY dist
LIMIT 3;

-- With ST_DWithin()
SELECT pointB.geom, pointB.id, ST_Distance(pointA.geom, pointB.geom) dist
FROM c_random_points_fm_1000_mtm7 pointA, k_random_points_fm_1000000_mtm7 pointB
WHERE pointA.id = 146 AND ST_DWithin(pointA.geom, pointB.geom, 10) -- 148
ORDER BY dist
LIMIT 3;

-- Find the biggest third smallest distance. Indetermined time.
SELECT pointA.*,(SELECT dist 
                 FROM (SELECT ST_Distance(pointA.geom, pointB.geom) dist
                       FROM k_random_points_fm_1000000_mtm7 pointB
                       ORDER BY dist ASC
                       LIMIT 4) foo
                 ORDER BY dist DESC
                 LIMIT 1)
FROM c_random_points_fm_1000_mtm7 pointA
ORDER BY dist DESC
LIMIT 1;

------------------------------------------------------
-- 4.2) Nearest neighbour - For 1 POINT from 1 000 000 POINTS. KNN method.
------------------------------------------------------
SELECT pointB.geom, pointB.id, ST_Distance(pointA.geom, pointB.geom) dist
FROM c_random_points_fm_1000_mtm7 pointA, k_random_points_fm_1000000_mtm7 pointB
WHERE pointA.id = 148
ORDER BY (SELECT geom FROM c_random_points_fm_1000_mtm7 WHERE id = 75) <-> pointB.geom
LIMIT 3;

------------------------------------------------------
-- 4.3) Nearest neighbour - For 1000 POINT from 1 000 000 POINTS. KNN method.
------------------------------------------------------
CREATE TABLE k_random_points_fm_3nearest_1000_mtm7 AS
SELECT pointA.id, pointA.geom, pointB.id id2, ST_Distance(pointA.geom, pointB.geom) dist
FROM c_random_points_fm_1000_mtm7 pointA, k_random_points_fm_1000000_mtm7 pointB
WHERE pointB.id = ANY ((SELECT array(SELECT id FROM k_random_points_fm_1000000_mtm7 
                                     ORDER BY geom <-> pointA.geom
                                     LIMIT 3))::integer[])
ORDER BY pointA.id, dist;

-- Display as table
SELECT * FROM k_random_points_fm_3nearest_1000_mtm7;

-- Display In OpenJump
SELECT * FROM c_random_points_fm_1000_mtm7;

------------------------------------------------------
-- 4.4) Nearest neighbour - For 1000 POINT from 6200 POLYLINES. KNN method.
------------------------------------------------------
-- Display all the routes in OpenJump
SELECT * FROM l_routes_fm_mtm7;

-- Find the first nearest. Actually find the nearest bounding box. 
SELECT point.id, point.geom, line.id id2, ST_Distance(point.geom, line.geom) dist
FROM c_random_points_fm_1000_mtm7 point, l_routes_fm_mtm7 line
WHERE line.id = ANY ((SELECT array(SELECT id FROM l_routes_fm_mtm7 
                                   ORDER BY geom <#> point.geom
                                   LIMIT 1))::integer[])
ORDER BY point.id, dist;

-- Display the bounding box for each route. Look at point 684.
SELECT id, ST_Envelope(geom) geom FROM l_routes_fm_mtm7;

-- Find the first nearest. Two step approach.
WITH first100 AS (
  SELECT point.id, point.geom, line.id id2, ST_Distance(point.geom, line.geom) dist
  FROM c_random_points_fm_1000_mtm7 point, l_routes_fm_mtm7 line
  WHERE line.id = ANY ((SELECT array(SELECT id FROM l_routes_fm_mtm7 
                                     ORDER BY geom <#> point.geom
                                     LIMIT 30))::integer[])
  ORDER BY point.id, dist
),
ordered AS (
SELECT *, ROW_NUMBER() OVER (PARTITION BY id ORDER BY dist) rownum FROM first100
)
SELECT * FROM ordered WHERE rownum < 2;

------------------------------------------------------
-- 5) MapAlgebra - Aggregate RASTERS together.
------------------------------------------------------
-- Display the Forest Montmorency in OpenJump in WGS84
SELECT ST_Transform(geom, 4326) geom FROM b_limites_fm_mtm7;

-- Create a buffered version of the Forest Monmorency limits
CREATE TABLE m_limites_fm_200m_mtm7 AS
SELECT ST_Buffer(geom, 200) geom
FROM b_limites_fm_mtm7;

-- Display
SELECT ST_Transform(geom, 4326) geom FROM m_limites_fm_200m_mtm7;

-- Display tiles intersecting with Foret Montmorency
SELECT rast::geometry geom
FROM e_elevation_fm_10x10_wgs84, m_limites_fm_200m_mtm7
WHERE ST_Intersects(rast, ST_Transform(geom, 4326));

-- Merge them into a one row raster.
CREATE TABLE n_elevation_fm_wgs84 AS
SELECT ST_Union(rast) rast
FROM e_elevation_fm_10x10_wgs84, m_limites_fm_200m_mtm7
WHERE ST_Intersects(rast, ST_Transform(geom, 4326));

-- Display the extent of the new raster 
SELECT rast::geometry FROM n_elevation_fm_wgs84;

-- Display avectorization of the new raster 
SELECT (ST_DumpAsPolygons(rast)).* FROM n_elevation_fm_wgs84;

------------------------------------------------------
-- 5.1) MapAlgebra - Compute HillShade.
------------------------------------------------------
-- Reproject the unioned raster
CREATE TABLE n_elevation_fm_mtm7 AS
SELECT ST_Transform(rast, 32187) rast
FROM n_elevation_fm_wgs84;

-- Display SRDI
SELECT ST_SRID(rast) srid FROM n_elevation_fm_mtm7;

-- Display 
SELECT rast::geometry FROM n_elevation_fm_mtm7;
SELECT * FROM b_limites_fm_mtm7;
SELECT (ST_DumpAsPolygons(rast)).* FROM n_elevation_fm_mtm7;

-- Compute the hillshade raster. 5s
CREATE TABLE n_hillshade_fm_mtm7 AS
SELECT ST_HillShade(rast, 1, '32BF', 180) rast
FROM n_elevation_fm_mtm7;

-- Display in OpenJump
SELECT (ST_DumpAsPolygons(rast)).* FROM n_hillshade_fm_mtm7;

------------------------------------------------------
-- 5.2) MapAlgebra - Reclass hillshade into 20 classes.
------------------------------------------------------
SELECT (ST_SummaryStats(rast)).* FROM n_hillshade_fm_mtm7;

CREATE TABLE n_hillshade_fm_20_mtm7 AS
SELECT ST_Reclass(rast, ROW(1, '0-150:0-10, (150-254: 10-20', '8BUI', 255)::reclassarg) rast
FROM n_hillshade_fm_mtm7;

SELECT (ST_SummaryStats(rast)).* FROM n_hillshade_fm_20_mtm7;

-- Display in OpenJump
SELECT (ST_DumpAsPolygons(rast)).* FROM n_hillshade_fm_20_mtm7;

------------------------------------------------------
-- 5.3) MapAlgebra - Same Reclass but using ST_Mapalgebra directly
------------------------------------------------------
CREATE TABLE n_hillshade_fm_20_mtm7_ma AS
SELECT ST_SetBandNodataValue(ST_MapAlgebra(rast, '8BUI', 
   'CASE
      WHEN 0 <= [rast] AND [rast] <= 150 THEN round(10 * [rast] / 150.0)
      WHEN 150 < [rast] AND [rast] <= 254 THEN 10 + round(10 * ([rast] - 150)/(254 - 150))
      ELSE 255
    END', 255), 255) rast
FROM n_hillshade_fm_mtm7;

SELECT (ST_Metadata(rast)).*, (ST_BandMetadata(rast)).* FROM n_hillshade_fm_mtm7
UNION ALL
SELECT (ST_Metadata(rast)).*, (ST_BandMetadata(rast)).* FROM n_hillshade_fm_20_mtm7
UNION ALL
SELECT (ST_Metadata(rast)).*, (ST_BandMetadata(rast)).* FROM n_hillshade_fm_20_mtm7_ma;

SELECT (ST_SummaryStats(rast)).* FROM n_hillshade_fm_mtm7
UNION ALL
SELECT (ST_SummaryStats(rast)).* FROM n_hillshade_fm_20_mtm7
UNION ALL
SELECT (ST_SummaryStats(rast)).* FROM n_hillshade_fm_20_mtm7_ma;

-- Display in OpenJump
SELECT (ST_DumpAsPolygons(rast)).* FROM n_hillshade_fm_20_mtm7_ma;

------------------------------------------------------
-- 6.1) Rasterization - POLYGON to RASTER. ST_AsRaster(), ST_Union(), ST_MapAlgebra() method.
------------------------------------------------------
-- Create a 10x10 tiled version of n_elevation_fm_mtm7 limited to the forest extent
CREATE TABLE o_elevation_fm_10x10_mtm7 AS
SELECT rast
FROM (SELECT ST_Tile(rast, 10, 10) rast FROM n_elevation_fm_mtm7) foo,
     m_limites_fm_200m_mtm7
WHERE ST_Intersects(rast, geom);

-- Index it.
CREATE INDEX o_elevation_fm_10x10_mtm7_rast_gist ON o_elevation_fm_10x10_mtm7 USING gist (st_convexhull(rast));

-- Add an id.
SELECT ST_AddUniqueID('o_elevation_fm_10x10_mtm7', 'rid', true);

-- Display 
SELECT rast::geometry FROM o_elevation_fm_10x10_mtm7;
SELECT (ST_DumpAsPolygons(rast)).* FROM o_elevation_fm_10x10_mtm7;

-- Rasterize the forest cover height. 20s
CREATE TABLE p_forestheight_fm_10x10_mtm7 AS
WITH forestrast AS (
  SELECT rid, ST_MapAlgebra(
                ST_Union(ST_AsRaster(geom, rast, '16BSI', COALESCE(cha_co::int, 0), -9999)),
                ST_AddBand(ST_MakeEmptyRaster(rast), '16BSI'::text, -9999, -9999), 
                '[rast1]', '16BSI', 'SECOND', NULL, '[rast1]'
                ) rast
  FROM a_forestcover_mtm7, o_elevation_fm_10x10_mtm7
  WHERE ST_Intersects(geom, rast)
  GROUP BY rid, rast
)
SELECT a.rid,
       CASE
         WHEN b.rid IS NULL THEN ST_AddBand(ST_MakeEmptyRaster(a.rast), '16BUI'::text, -9999, -9999)
         ELSE b.rast
       END rast
FROM o_elevation_fm_10x10_mtm7 a LEFT OUTER JOIN forestrast b 
ON a.rid = b.rid;

-- Display tiles in OpenJump
SELECT rast::geometry FROM p_forestheight_fm_10x10_mtm7;

-- Display pixels in OpenJump
SELECT (ST_DumpAsPolygons(rast)).* FROM p_forestheight_fm_10x10_mtm7;

------------------------------------------------------
-- 6.2) Rasterization - POLYGON to RASTER. PostGIS Addons method.
------------------------------------------------------
-- Make a table with an integer
CREATE TABLE q_forestcover_int_mtm7 AS
SELECT COALESCE(cha_co::int, 0) cha_co, geom
FROM a_forestcover_mtm7;

-- Make sure there is a spatial index on the table to rasterize
CREATE INDEX q_forestcover_int_mtm7_geom_gist ON q_forestcover_int_mtm7 USING gist (geom);

-- Rasterize using PostGIS Addons ST_ExtractToRaster(). 40s
CREATE TABLE r_forestheight_fm_10x10_mergedbiggest_mtm7 AS
SELECT ST_ExtractToRaster( 
                    ST_AddBand(
                        ST_MakeEmptyRaster(rast), '32BF'::text, -9999, -9999), 
                    'public', 
                    'q_forestcover_int_mtm7', 
                    'geom', 
                    'cha_co', 
                    'VALUE_OF_MERGED_BIGGEST'
                  ) rast 
FROM o_elevation_fm_10x10_mtm7;

-- Display pixels in OpenJump
SELECT (ST_DumpAsPolygons(rast)).* FROM r_forestheight_fm_10x10_mergedbiggest_mtm7;
------------------------------------------------------
-- 5.4) MapAlgebra - Add a RASTER to another RASTER.
------------------------------------------------------
-- Reclass the height from a code to meters
CREATE TABLE s_forestheight_meter_fm_10x10_mtm7 AS
SELECT ST_Reclass(rast, ROW(1, '0:0, 1:22, 2:19, 3:14, 4:9, 5:5, 6:3', '8BUI', 255)::reclassarg) rast
FROM p_forestheight_fm_10x10_mtm7;

-- Display pixels in OpenJump
SELECT (ST_DumpAsPolygons(rast)).* FROM s_forestheight_meter_fm_10x10_mtm7;

-- Add the two height together only when the tiles correspond
CREATE TABLE s_canopyheight_10x10_mtm7 AS
SELECT ST_MapAlgebra(e.rast, fc.rast, '[rast1] + [rast2]', NULL,'INTERSECTION') rast
FROM o_elevation_fm_10x10_mtm7 e, s_forestheight_meter_fm_10x10_mtm7 fc
WHERE ST_UpperLeftX(e.rast) = ST_UpperLeftX(fc.rast) AND ST_UpperLeftY(e.rast) = ST_UpperLeftY(fc.rast);

-- Display pixels in OpenJump
SELECT (ST_DumpAsPolygons(rast)).* FROM s_canopyheight_10x10_mtm7;

------------------------------------------------------
-- 7) Overlap removal
------------------------------------------------------
-- Add our own unique id to the forest cover.
SELECT ST_AddUniqueID('a_forestcover_mtm7', 'id', true);

-- Identify overlapping parts. 25s
CREATE TABLE t_forest_overlaping_parts_mtm7 AS
SELECT a.id, ST_Intersection(a.geom, b.geom) geom, ST_Area(ST_Intersection(a.geom, b.geom)) area
FROM a_forestcover_mtm7 a, 
     a_forestcover_mtm7 b
WHERE a.id < b.id AND (ST_Overlaps(a.geom, b.geom) OR ST_Contains(a.geom, b.geom) OR ST_Contains(b.geom, a.geom))
ORDER BY area DESC;

-- Display
SELECT * FROM t_forest_overlaping_parts_mtm7;

-- Alternative way to check for overlaps. 24s
SELECT 'sum'::text, sum(ST_Area(geom)) area FROM a_forestcover_mtm7
UNION ALL
SELECT 'union'::text, ST_Area(ST_Union(geom)) area FROM a_forestcover_mtm7

-- Create a new table without overlaps. 67s
CREATE TABLE t_forest_without_overlaping_parts_mtm7 AS 
SELECT a.id, ST_DifferenceAgg(a.geom, b.geom) geom
FROM a_forestcover_mtm7 a, 
     a_forestcover_mtm7 b
WHERE ST_Equals(a.geom, b.geom) OR 
      ((ST_Contains(a.geom, b.geom) OR 
        ST_Contains(b.geom, a.geom) OR 
        ST_Overlaps(a.geom, b.geom)) AND 
       (ST_Area(a.geom) < ST_Area(b.geom) OR 
        (ST_Area(a.geom) = ST_Area(b.geom) AND 
         ST_AsText(a.geom) < ST_AsText(b.geom))))
GROUP BY a.id;

-- Check areas. 24s
SELECT 'sum'::text, sum(ST_Area(geom)) area FROM t_forest_without_overlaping_parts_mtm7
UNION ALL
SELECT 'union'::text, ST_Area(ST_Union(geom)) area FROM t_forest_without_overlaping_parts_mtm7

-- Display
SELECT * FROM t_forest_without_overlaping_parts_mtm7;

-- Try to find overlapping part again. 214s
CREATE TABLE t_forest_without_overlaping_parts_overlaping_parts_mtm7 AS
SELECT a.id, ST_Intersection(a.geom, b.geom) geom, ST_Area(ST_Intersection(a.geom, b.geom)) area
FROM t_forest_without_overlaping_parts_mtm7 a, 
     t_forest_without_overlaping_parts_mtm7 b
WHERE a.id < b.id AND (ST_Overlaps(a.geom, b.geom) OR ST_Contains(a.geom, b.geom) OR ST_Contains(b.geom, a.geom))
ORDER BY area DESC;

-- Display
SELECT * FROM t_forest_without_overlaping_parts_overlaping_parts_mtm7;